import numpy as np
import matplotlib.pyplot as plt

def plot_interactions(phi_map,x,y):
    values = np.around([[phi_map[(i,j)].item() for i in range(len(x))] for j in range(len(y))], decimals=2)
    fig = plt.figure()
    ax = plt.axes()
    im = ax.imshow(values, cmap=plt.get_cmap('Reds'))

    ax.set_xticks(np.arange(len(x)))
    ax.set_yticks(np.arange(len(y)))
    ax.set_xticklabels(x, fontsize=11)
    ax.set_yticklabels(y, fontsize=11)
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")
    for i in range(len(y)):
        for j in range(len(x)):
            text = ax.text(j, i, values[i, j],
                           ha="center", va="center", color="w")
    return fig, ax



import torch
import warnings
import math
import random
import numpy as np
import pdb

import scipy as sp
import sklearn
from transformers import BartTokenizer, BartForConditionalGeneration
import torch

def binomial_coef_dist(n):
    dist = [math.comb(n, i+1) for i in range(n//2)]
    total = sum(dist)
    dist = [density / total for density in dist]
    return dist, total

class BasicPM():
    def __init__(self):
        pass
    def perturb_inputs(self, x, num=1):
        """
        argument:
            x: the tokenized input sentence
        return:
            x_set: the perturbed xs
            z_set: the simplified features of x_set, {0,1}^|x|, tensor
        """
        if num != 1:
            warnings.warn("BasicPM will always set argument num == 1")
        return [x],torch.tensor([[1.0 for tok in x]])


class RandomPM(BasicPM):
    """
    randomly choose tokens to be replaced with sub_t
    """
    def __init__(self, sub_t="", denoising=False):
        super().__init__()
        self.sub_t = sub_t
        self.denoising = denoising
        if self.denoising:
            self.sub_t = '<mask>'
            self.bart_tokenizer = BartTokenizer.from_pretrained('facebook/bart-base')
            self.bart_model = BartForConditionalGeneration.from_pretrained('facebook/bart-base').to('cuda')

    def _select_repl_num(self, dist_scale):
        """
        select the number of tokens to be replaces #repl_num, following binomial distribution
        """
        pos = random.random()
        for i in range(len(dist_scale)):
            if pos < dist_scale[i]:
                break
        return i+1

    def _denoise_x_set(self, x_set):
        inputs = self.bart_tokenizer(x_set, max_length=256, return_tensors='pt', padding=True).to('cuda')
        summary_ids = self.bart_model.generate(
            inputs['input_ids'],
            top_k=10, top_p=0.9, temperature=0.9, max_length=256,
            early_stopping=True, num_return_sequences=1
        )
        lines = [self.bart_tokenizer.decode(g, skip_special_tokens=True, clean_up_tokenization_spaces=False) for g in summary_ids]
        x_set = lines
        return x_set

    def perturb_inputs(self, x, num=1000, with_i=None):
        """
        allow half tokens at most can be replaced
        """
        dist, num_comb = binomial_coef_dist(len(x))
        dist_scale = [sum(dist[:i+1]) for i in range(len(dist))]
        num = num if num < num_comb*4 else num_comb*4

        """
        choose tokens to be replaced with sub_t
        """
        if with_i is None:
            x_set, z_set = [], []
        else:
            x_set, x_set_with_i = [], []
            weights = []
        for _ in range(num):
            repl_num = self._select_repl_num(dist_scale)
            x_set.append(list(x))
            if with_i is None:
                z_set.append(np.ones((len(x),)))
                repl_list = random.sample(list(range(len(x))), repl_num)
                for t in repl_list:
                    x_set[-1][t] = self.sub_t
                    z_set[-1][t] = 0.
            else:
                x_set_with_i.append(list(x))
                indices_to_repl = list(range(len(x)))
                indices_to_repl.remove(with_i)
                repl_list = random.sample(indices_to_repl, repl_num)
                for t in repl_list:
                    x_set[-1][t] = self.sub_t
                    x_set_with_i[-1][t] = self.sub_t
                x_set[-1][with_i] = self.sub_t
                weights.append(1/(dist[repl_num-1]*len(x)))
        if self.denoising:
            x_set = self._denoise_x_set([' '.join(x) for x in x_set])
        if with_i is None:
            return x_set, torch.tensor(z_set, dtype=torch.float32), x
        else:
            return x_set, x_set_with_i, weights, x


class LIMERandomPM(RandomPM):
    def perturb_inputs(self, x, num=1000):
        """
        allow half tokens at most can be replaced
        choose tokens to be replaced with sub_t
        """
        dist, num_comb = binomial_coef_dist(len(x))
        num = num if num < num_comb*4 else num_comb*4

        x_set, z_set = [], []

        sample = np.random.randint(1,len(x)//2+1,num-1)
        
        x_set.append(list(x))
        z_set.append(np.ones((len(x),)))
        for i in range(num-1):
            repl_num = sample[i]
            x_set.append(list(x))
            z_set.append(np.ones((len(x),)))
            repl_list = random.sample(list(range(len(x))), repl_num)
            for t in repl_list:
                x_set[-1][t] = self.sub_t
                z_set[-1][t] = 0.

        if self.denoising:
            x_set = self._denoise_x_set([' '.join(x) for x in x_set])
        return x_set, torch.tensor(z_set, dtype=torch.float32), x



import torch
import numpy as np
import random
import pdb

from scipy.stats import skew
from collections import Counter

def get_expl(x, expl, ratio=0.2, remain_masks=False):
    if expl is None:
        x_entities = [tok for tok in x if random.random() < ratio] if not remain_masks else [tok if random.random() < ratio else "__" for tok in x ]
    else:
        k = int(len(x) * ratio // 1)
        topk = torch.topk(expl, k) if not remain_masks else torch.topk(expl, max(k,1))
        x_entities = [tok for ind, tok in enumerate(x) if ind in topk.indices] if not remain_masks else [tok if ind in topk.indices else "__" for ind, tok in enumerate(x)]
    if remain_masks:
        merged = [x_entities[0]]
        for tok in x_entities[1:]:
            if tok == "__" and merged[-1] == "__":
                continue
            else:
                merged.append(tok)
        x_entities = merged
    return x_entities

def remove_expl(x, expl, ratio=0.2):
    """
    remove given explanation from x
    if None explanation is given, randomly remove
    """
    if expl is None:
        x_re = [tok for tok in x if random.random() >= ratio]
    else:
        k = int(len(x) * ratio // 1)
        topk = torch.topk(expl, k)
        x_re = [tok for ind, tok in enumerate(x) if ind not in topk.indices]
    return x_re

def get_ppl(probs, y, y_inds=None):
    if y_inds is None:
        ent = np.sum(np.log(p[yi]) for p, yi in zip(probs[0], y)) / len(y)
    else:
        ent = np.sum(np.log(p[yi]) for i, (p, yi) in enumerate(zip(probs[0], y)) if i in y_inds) / len(y_inds)
    return np.exp(-ent), ent

def ppl_c_add(expl, x, y, model_f, ratio=0.2):
    """
    additive perplexity changes
    """
    x_add = get_expl(x, expl, ratio=ratio)
    y_probs_add, y_inds = model_f([x_add], label=y, is_x_tokenized=True, is_y_tokenized=True)
    ppl_add, ent_add = get_ppl(y_probs_add.cpu(), y_inds)
    return ent_add, ppl_add, x_add

def ppl_c(expl, x, y, model_f, ratio=0.2):
    """
    perplexity changes
    """
    x_re = remove_expl(x, expl, ratio=ratio)

    y_probs, y_inds = model_f([x], label=y, is_x_tokenized=True, is_y_tokenized=True)
    y_probs_re, _ = model_f([x_re], label=y, is_x_tokenized=True, is_y_tokenized=True)

    ppl, ent = get_ppl(y_probs.cpu(), y_inds)
    ppl_re, ent_re = get_ppl(y_probs_re.cpu(), y_inds)

    entc = ent_re - ent
    return entc, x_re, ppl, ppl_re



import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import pdb

from sklearn.metrics.pairwise import pairwise_distances
from sklearn.linear_model import Ridge
import numpy as np
import random

class Explainer():
    """
    The base class for various explainers
    arguments:
        model_f: the tested model function, who generates y given x
            require the outputs in the form (probabilities of output sequence, y's tokens ids), having the same length
        x: the input
        y: the generated sequence
    return:
        phi_set: correspond to the weight of each x_i
    """
    def __init__(self, model_f, x, y):
        self.phi_map = {}
        self.model_f = model_f
        self.x = x
        self.y = y
        
    def get_prob(self, probs, y):
        y_probs = [[p[yi] for p, yi in zip(prob, y)] for prob in probs]
        return y_probs


class LERG(Explainer):
    """
    The base class for all Local Explanation methods for Response Generation
    """
    def __init__(self, model_f, x, y, perturb_f, tokenizer, max_iters=50, device="cuda" if torch.cuda.is_available() else "cpu"):
        super().__init__(model_f, x, y)
        self.perturb_inputs = perturb_f
        self.max_iters = max_iters
        self.tokenizer = tokenizer
        self.device = device

    def combine_sequence(self, phi_sets):
        """
        phi_sets.shape: (output_dim) x (input_dim)
        """
        return torch.sum(phi_sets,dim=0)

    def map_to_interactions(self, phi_sets):
        phi_map = {}
        for yi in range(phi_sets.shape[0]):
            for xi in range(phi_sets.shape[1]):
                phi_map[(xi,yi)] = phi_sets[yi,xi]
        return phi_map

    def save_exp(self, save_path):
        if self.phi_set is not None:
            torch.save([self.phi_set, self.phi_map, self.components, self.y], save_path)
        else:
            raise ValueError("run get_local_exp() first")


class LERG_LIME(LERG):
    """
    LERG by LIME
    """
    def __init__(self, model_f, x, y, perturb_f, tokenizer, max_iters=50, device="cuda" if torch.cuda.is_available() else "cpu"):
        super().__init__(model_f, x, y, perturb_f, tokenizer, max_iters, device=device)
        self.batchsize = 64

    def get_local_exp(self):
        self.x = self.tokenizer.tokenize(self.x)
        self.y = self.tokenizer.tokenize(self.y)

        x_set, z_set, self.components = self.perturb_inputs(self.x)

        y_probs = []
        for i in range(len(x_set)//self.batchsize + 1 if len(x_set)%self.batchsize > 0 else 0):
            probs,y = self.model_f(x_set[i*self.batchsize:(i+1)*self.batchsize], label=self.y, is_x_tokenized=True, is_y_tokenized=True)
            y_probs_batch = self.get_prob(probs, y)
            y_probs_batch = torch.tensor(y_probs_batch)
            y_probs.append(y_probs_batch)
        y_probs = torch.cat(y_probs,dim=0)

        D = pairwise_distances(z_set,z_set[0].view(1,-1),metric='cosine')
        kernel_width = 25# as LIME's original implementation
        weights = torch.tensor(np.sqrt(np.exp(-(D ** 2) / kernel_width ** 2)), requires_grad=False).to(self.device)

        self.expl_model = nn.Linear(z_set.shape[1],len(y),bias=False).to(self.device)
        self.optimizer = torch.optim.SGD(self.expl_model.parameters(), lr=5e-1)

        for i in range(self.max_iters):
            for z_batch, y_probs_batch, w_batch in zip(torch.split(z_set, self.batchsize), torch.split(y_probs, self.batchsize), torch.split(weights,self.batchsize)):
                z_batch = z_batch.to(self.device)
                y_probs_batch = y_probs_batch.to(self.device)
                preds = self.expl_model(z_batch)# the original version for classifier
                loss = torch.mean(w_batch * (preds - y_probs_batch) ** 2)
                loss.backward()
                self.optimizer.step()
                self.optimizer.zero_grad()

        with torch.no_grad():
            phi_sets = self.expl_model.weight
            self.phi_set = self.combine_sequence(phi_sets)
            self.phi_map = self.map_to_interactions(phi_sets)

        return self.phi_set, self.phi_map, self.components, self.y


class LERG_R(LERG):
    """
    LERG use ratio probability
    """
    def __init__(self, model_f, x, y, perturb_f, tokenizer, max_iters=50, device="cuda" if torch.cuda.is_available() else "cpu"):
        super().__init__(model_f, x, y, perturb_f, tokenizer, max_iters, device=device)
        self.batchsize = 64

    def get_local_exp(self):
        self.x = self.tokenizer.tokenize(self.x)
        self.y = self.tokenizer.tokenize(self.y)
        x_set, z_set, self.components = self.perturb_inputs(self.x)

        gold_probs,y = self.model_f([self.x], label=self.y, is_x_tokenized=True, is_y_tokenized=True)
        gold_probs = self.get_prob(gold_probs, y)
        gold_probs = torch.tensor(gold_probs)
        gold_probs = gold_probs[0]

        probs,y = self.model_f(x_set, label=self.y, is_x_tokenized=True, is_y_tokenized=True)
        y_probs = self.get_prob(probs, y)
        y_probs = torch.tensor(y_probs)
        y_probs /= gold_probs

        self.expl_model = nn.Linear(z_set.shape[1],len(y),bias=False).to(self.device)
        self.optimizer = torch.optim.SGD(self.expl_model.parameters(), lr=5e-1)

        for i in range(self.max_iters):
            for z_batch, y_probs_batch in zip(torch.split(z_set, self.batchsize), torch.split(y_probs, self.batchsize)):
                z_batch = z_batch.to(self.device)
                y_probs_batch = y_probs_batch.to(self.device)
                preds = self.expl_model(z_batch)
                loss = F.mse_loss(preds,y_probs_batch)
                loss.backward()
                self.optimizer.step()
                self.optimizer.zero_grad()

        with torch.no_grad():
            phi_sets = self.expl_model.weight
            self.phi_set = self.combine_sequence(phi_sets)
            self.phi_map = self.map_to_interactions(phi_sets)

        return self.phi_set, self.phi_map, self.components, self.y


class LERG_SHAP(LERG):
    """
    LERG use SampleShapley (original)
    """
    def __init__(self, model_f, x, y, perturb_f, tokenizer, device="cuda" if torch.cuda.is_available() else "cpu"):
        super().__init__(model_f, x, y, perturb_f, tokenizer, max_iters=0, device=device)

    def get_local_exp(self):
        self.x = self.tokenizer.tokenize(self.x)
        self.y = self.tokenizer.tokenize(self.y)

        phi_sets = []
        for i in range(len(self.x)):
            x_set, x_set_with_i, weights, self.components = \
                self.perturb_inputs(self.x, num=500//len(self.x), with_i=i)# results in total 1000 samples as LERG_LIME

            probs,y = self.model_f(x_set, label=self.y, is_x_tokenized=True, is_y_tokenized=True)
            y_probs = self.get_prob(probs, y)
            y_probs = torch.tensor(y_probs)

            probs, _ = self.model_f(x_set_with_i, label=self.y, is_x_tokenized=True, is_y_tokenized=True)
            y_probs_with_i = self.get_prob(probs, y)
            y_probs_with_i = torch.tensor(y_probs_with_i)
            weights = torch.tensor(weights).view(-1,1)
            phi_sets.append(torch.mean((y_probs_with_i - y_probs)*weights, dim=0))

        phi_sets = torch.stack(phi_sets).transpose(0,1)
        self.phi_set = self.combine_sequence(phi_sets)
        self.phi_map = self.map_to_interactions(phi_sets)

        return self.phi_set, self.phi_map, self.components, self.y


class LERG_SHAP_log(LERG):
    """
    LERG use Shapley value with sample mean (Logarithm)
    """
    def __init__(self, model_f, x, y, perturb_f, tokenizer, device="cuda" if torch.cuda.is_available() else "cpu"):
        super().__init__(model_f, x, y, perturb_f, tokenizer, max_iters=0, device=device)

    def get_local_exp(self):
        self.x = self.tokenizer.tokenize(self.x)
        self.y = self.tokenizer.tokenize(self.y)

        phi_sets = []
        for i in range(len(self.x)):
            x_set, x_set_with_i, weights, self.components = \
                self.perturb_inputs(self.x, num=500//len(self.x), with_i=i)# results in total 1000 samples as LERG_LIME

            probs,y = self.model_f(x_set, label=self.y, is_x_tokenized=True, is_y_tokenized=True)
            y_probs = self.get_prob(probs, y)
            y_probs = torch.tensor(y_probs)

            probs, _ = self.model_f(x_set_with_i, label=self.y, is_x_tokenized=True, is_y_tokenized=True)
            y_probs_with_i = self.get_prob(probs, y)
            y_probs_with_i = torch.tensor(y_probs_with_i)
            phi_sets.append(torch.mean((torch.log(y_probs_with_i) - torch.log(y_probs)), dim=0))

        phi_sets = torch.stack(phi_sets).transpose(0,1)
        self.phi_set = self.combine_sequence(phi_sets)
        self.phi_map = self.map_to_interactions(phi_sets)

        return self.phi_set, self.phi_map, self.components, self.y
